/*
  Tool            sort_bug
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Windows GUI TSE 4.50 rc 15
  Version         1   23 Dec 2023

  This tool inconsistently reproduces the bug that the internal Sort() aborts
  with the "I'm sorry, Dave." memory error.

  These are factors that seem to enhance the chance to reproduce the memory
  abort:
    That this macro does a GetGlobalStr().
    Starting the macro ASAP after starting the editor.
      But not from the command line.
    Starting the macro "from the compiler".
      I think this "helps" because compile.s uses global vars too.

  These are factors that might NOT help to reproduce the memory abort:
    Once the macro does not abort, it is unlikely to abort in a next try in the
    same TSE session.
      This macro still has the cability to do multiple "runs", but testing
      indicated that this did not at all help to reproduce the abort.

  Note:
    This macro writes logging to TSE's TMP directory as
      tse_log_<macro name>_<time in centiseconds since midnight>.log

  Back story
    I occasionally got this memory abort in my personal TSE installation since
    starting to seriously test TSE's sorts.
    That must be since at least a month, so the sort changes in TSE 4.50 rc 15
    are not suspect.
    The first version of this tool was able to reproduce the abort frequently
    in my personal TSE installation but not in a new TSE installation.
    I have tons of macros autoloaded, so they were suspect.
    Purging all my loaded macros after starting TSE did not solve the abort,
    leading to my conclusion that they already did lasting damage somehow.
    So I one by one deleted macros from the autoload list, restarted TSE,
    and retested.
    Because the bug must have something to do with memory I became suspicious
    of global variables, and started deleting macros that used them next.
    As soon as I deleted the last macro that uses them from my autoload list
    this tool no longer reproduced the bug in my TSE installation.
    Surprisingly that last macro only uses a GetGlobalStr(), not a
    SetGlobalSTR, and that turned out to be relevant.
    I added a GetGlobalStr() to this tool, and now it occasionally reproduces
    the abort in a new TSE installation too. Way less, unfortunately, but given
    the earlier listed "enhancers" often enough to prove that there is a bug in
    TSE itself.
    Equally unfortunately this means there still is an unknown seemingly random
    factor involved in aborting or not aborting.

*/





// Start of zz_log implementation

integer zz_log_handle = -1

integer proc zz_log_write(string msg)
  integer old_DateFormat = Set(DateFormat, 6)
  integer old_timeFormat = Set(TimeFormat, 1)
  integer ok             = TRUE

  #ifdef LINUX
    string NEWLINE [1] = Chr(10)
  #else
    string NEWLINE [2] = Chr(13) + Chr(10)
  #endif

  if zz_log_handle == -1
    Warn('No open log to write to.')
    ok = FALSE
  else
    if fWrite(zz_log_handle, GetDateStr() + ' ' + GetTimeStr() + ' ') == -1
      Warn('Log write date time prefix failed.')
      ok = FALSE
    else
      if fWrite(zz_log_handle, msg) == -1
        Warn('Log write message failed.')
        ok = FALSE
      else
        if fWrite(zz_log_handle, NEWLINE) == -1
          Warn('Log write newline failed.')
          ok = FALSE
        endif
      endif
    endif
  endif
  Set(DateFormat, old_DateFormat)
  Set(TimeFormat, old_timeFormat)
  return(ok)
end zz_log_write

integer proc zz_log_open()
  string  log_fqn [MAXSTRINGLEN] = 'tse_log_' +
                                   SplitPath(CurrMacroFilename(), _NAME_) +
                                   '_' + Format(GetTime():7:'0') + '.log'
  integer log_attrs              = 0
  integer ok                     = TRUE

  #ifdef LINUX
    log_fqn = '/tmp/' + log_fqn
  #else
    log_fqn = GetEnvStr('TMP') + '\' + log_fqn
  #endif

  if zz_log_handle <> -1
    Warn('Log already open.')
  else
    log_attrs = FileExists(log_fqn)
    if log_attrs
      if log_attrs & _DIRECTORY_
        Warn('Cannot log to directory.')
        ok = FALSE
      else
        zz_log_handle = fOpen(log_fqn, _OPEN_READWRITE_)
        if zz_log_handle == -1
          Warn('Open log failed.')
          ok = FALSE
        else
          if fSeek(zz_log_handle, 0, _SEEK_END_) == -1
            Warn('First log fSeek failed.')
            ok = FALSE
          else
            zz_log_write('Log reopened.')
          endif
        endif
      endif
    else
      if FileExists(SplitPath(log_fqn, _DRIVE_|_PATH_)) & _DIRECTORY_
        zz_log_handle = fCreate(log_fqn)
        if zz_log_handle == -1
          Warn('Creating log failed.')
          ok = FALSE
        else
          ok = zz_log_write('Log opened.')
        endif
      else
        Warn('Cannot log to non-existing directory.')
        ok = FALSE
      endif
    endif
  endif

  return(ok)
end zz_log_open

integer proc zz_log_close()
  integer ok = FALSE
  if zz_log_handle <> -1
    ok = zz_log_write('Log closed.')
    if fClose(zz_log_handle)
      zz_log_handle = -1
    else
      Warn('Failed to close log.')
      ok = FALSE
    endif
  endif
  return(ok)
end zz_log_close

// End of zz_log implementation.




// Global constants and semi-constants.
string MACRO_NAME [MAXSTRINGLEN] = ''

#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif

#if INTERNAL_VERSION > 12375
  #define IGNORE_CASE _IGNORE_CASE_
#else
  #define IGNORE_CASE 1
#endif


// For debugging:
proc log_current_buffer_to_tmp_dir(string identifier)
  integer time = GetTime()
  zz_log_write('B log_current_buffer_to_tmp_dir')
  SaveAs(Format(AddTrailingSlash(GetEnvStr('TMP')),
                time, '_', identifier, '.txt'),
         _DONT_PROMPT_|_OVERWRITE_)
  while time == GetTime()
    Delay(1)
  endwhile
  zz_log_write('E log_current_buffer_to_tmp_dir')
end log_current_buffer_to_tmp_dir


proc copy_buffer(integer from_id, integer to_id)
  GotoBufferId(from_id)
  MarkAll()
  GotoBufferId(to_id)
  SetUndoOff()
  EmptyBuffer()
  CopyBlock()
end copy_buffer


proc make_test_set_1(integer buffer_id)
  string  char [1] = ''
  integer col      = 0
  integer line     = 0

  GotoBufferId(buffer_id)
  SetUndoOff()
  EmptyBuffer()

  for line = 1 to 10000
    AddLine()
    EndLine()
    char = 'a'
    for col = 1 to 283 + 1 - (line mod 283 + 1)
      InsertText(char)
      if char == 'z'
        char = 'a'
      else
        char = Chr(Asc(char) + 1)
      endif
    endfor
  endfor
end make_test_set_1


proc WhenPurged()
  zz_log_write('B WhenPurged')
  zz_log_write('E WhenPurged')
  zz_log_close()
end WhenPurged


proc WhenLoaded()
  zz_log_open()
  zz_log_write('B WhenLoaded')
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)
  zz_log_write('E WhenLoaded')
end WhenLoaded


proc Main()
  integer old_MsgLevel     = Set(MsgLevel, _WARNINGS_ONLY_)
  integer run              = 0
  integer sort_id          = NewFile()
  integer unsorted_id      = NewFile()

  zz_log_write('B Main')

  if FALSE
    MACRO_NAME = MACRO_NAME
    log_current_buffer_to_tmp_dir('InternalSortBefore')
  endif

  make_test_set_1(unsorted_id)

  for run = 1 to 1
    KeyPressed()
    Message('Run'; run)
    KeyPressed()

    GetGlobalStr(Format('sort_bug:IAmGoingToTryToMakeTheInternalSortAbort'))

    copy_buffer(unsorted_id, sort_id)
    MarkAll()
    zz_log_write(Format(run:10; 'Just before Sort().'))
    Sort()
    zz_log_write(Format(run:10; 'Just after  Sort().'))
  endfor

  if FALSE
    log_current_buffer_to_tmp_dir('InternalSortAfter')
  endif

  PopLocation()
  AbandonFile(sort_id)
  AbandonFile(unsorted_id)
  Set(MsgLevel, old_MsgLevel)
  PurgeMacro(CurrMacroFilename())
  zz_log_write('E Main')
end Main

